from functools import wraps
import re
import datetime
from flask import current_app, request, g


class TokenMixin(object):
    def get_scopes(self):
        """Returns a list of scope strings that are authorised for the token."""
        raise NotImplementedError()

    def has_scope(self, scope):
        return scope in self.get_scopes()

    def serialize(self):
        """Serialize the token into a string.

        Your token class should implement this.
        """
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, app, token_str):
        """Deserialize a token into a token object.

        Your token class should implement this."""
        raise NotImplementedError


class JWTFactory(object):
    def __init__(self, signing_secret, expiry):
        self._signing_secret = signing_secret
        self._expiry = expiry

    def __call__(self):
        import jwt

        _scopes = self.scopes
        _expiry = self._expiry
        _signing_secret = self._signing_secret

        class JSONWebToken(TokenMixin):
            def __init__(self, scopes=None):
                self.scopes = scopes

            def get_scopes(self):
                """Returns a list of scope strings that are authorised for the token."""
                return self.scopes

            def deserialize_data(self, data):
                raise NotImplementedError()

            def serialize_data(self):
                raise NotImplementedError()

            def serialize():
                payload = {
                    'scopes': _scopes,
                    'exp': datetime.datetime.utcnow() + _expiry,
                    'data': self.serialize()
                }

                return jwt.encode(payload, _signing_secret).decode('UTF-8')

            @classmethod
            def deserialize(Cls, app, token_str):
                """Converts a JWT payload into a"""
                try:
                    decoded = jwt.decode(token_str, self._signing_secret)
                except jwt.InvalidTokenError:
                    raise LookupError("The provided JWT payload was invalid.")

                token = Cls(scopes=decoded.get('scopes', []))
                token.deserialize(data=decoded.get('data', {}))
                return token


bearer_re = re.compile(r'Bearer (.+)')

class TokenAuth():
    def __init__(self, app, token_cls):
        self._token_cls = token_cls
        self._app = app
        app.token_auth = self

    def _inject_token(self, _request):
        token = self._parse_request(_request)
        if token is None:
            raise Exception("The provided token was invalid.")

        try:
            g.token = self._token_cls.deserialize(self._app, token)
        except LookupError as e:
            raise Exception("The provided token was invalid. {}".format(e))

        return token

    def _scopes_accepted(self, _request, scopes, orig, fargs, fkwargs):
        """Given a flask request and list of accepted scopes, perform auth 
        and access control."""
        try:
            self._inject_token(_request)
        except Exception as e:
            return str(e), 403

        allowed_scopes = g.token.get_scopes()
        for scope in scopes:
            if scope in allowed_scopes:
                return orig(*fargs, **fkwargs)

        return "The provided token does not have an accepted scope. It needs "\
               "one of [{}], but only has [{}]".format(
                ', '.join(scopes),
                ', '.join(allowed_scopes))

    def _scopes_required(self, _request, scopes, orig, fargs, fkwargs):
        """Given a flask request and list of required scopes, perform 
        authentication and access control."""
        try:
            self._inject_token(_request)
        except Exception as e:
            return str(e), 403

        allowed_scopes = g.token.get_scopes()
        for scope in scopes:
            if scope not in allowed_scopes:
                return "The provided token does not have the required scopes "\
                       "[{}]. It has [{}]".format(
                        ', '.join(scopes),
                        ', '.join(allowed_scopes)), 403

        return orig(*fargs, **fkwargs)

    def _parse_request(self, _request):
        """Extract a token from a passed flask request"""
        auth_header = _request.headers.get('Authorization', None)
        if auth_header is not None:
            # Loading from the Auth header. Expect `Bearer <token>` format.
            if len(auth_header) > 0:
                result = bearer_re.search(auth_header)
                if result is not None:
                    return result.group(1)
        else:
            # Attempt to load from GET params
            return _request.args.get('token')


def scopes_required(*scopes):
    def _inner(f):
        @wraps(f)
        def _wrapped(*args, **kwargs):
            return current_app.token_auth._scopes_required(request, scopes, f, args, kwargs)
        return _wrapped
    return _inner

def scopes_accepted(*scopes):
    def _inner(f):
        @wraps(f)
        def _wrapped(*args, **kwargs):
            return current_app.token_auth._scopes_required(request, scopes, f, args, kwargs)
        return _wrapped
    return _inner



from flask import jsonify

def format_errors(*errors):
    return {
        '_errors': errors
    }

def format_error(error):
    return format_errors(error)


def output_json(data, code, headers=None):
    """
    ```
    api.representations['application/json'] = output_json
    ```
    
    Generates better formatted responses for RESTFul APIs. 

    If the restful resource responds with a string, with a non 200 error,
    the response will look like

    ```
    {
        "_errors": ["String the user responded with"]
    }
    ```

    Likewise, a 200 response will look like:

    ```
    {
        "message": "String the user returned with."
    }
    ```
    
    If a Non-200 response occured, and flask-restful added it's own error 
    message in the "message" field of the response data, this data is moved into
    "_errors":

    ```
    {
        "_errors": ["You don't have the permission to access the requested resource..."]
    }
    ```

    All data is returned using flask's jsonify method. This means you can 
    use simplejson to return decimal objects from your flask restful resources.
    """

    if type(data) is str:
        # Handle view returning a string.
        message = data
        if code != 200:
            data = format_errors(message)
        else:
            data = {'message': message}

    elif code != 200 and type(data) is dict and 'message' in data:
        # Flask-Restful returns non-200 error messages to the user.
        # Let's show them.
        data = format_errors(data.get('message'))

    resp = jsonify(data)

    if code:
        resp.status_code = code

    resp.headers.extend(headers)
    return resp



try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

def build_url(url, scheme=None, username=None, password=None, hostname=None, 
              port=None, path=None):
    dsn = urlparse(url)

    if scheme is None: scheme = dsn.scheme
    if username is None: username = dsn.username
    if password is None: password = dsn.password
    if hostname is None: hostname = dsn.hostname
    if port is None: port = dsn.port
    if path is None: path = dsn.path

    def build_auth():
        if username is not None or password is not None:
            return '{}:{}@'.format(username or '', password or '')
        return ''

    def build_port():
        if port is not None:
            return ':{}'.format(port)
        return ''

    return '{scheme}://{auth}{hostname}{port}{path}'.format(
        scheme=scheme,
        auth=build_auth(),
        hostname=hostname or '',
        port=build_port(),
        path=path
    )



from flask import request
from marshmallow import Schema, fields
from webargs import fields as wfields
from webargs.flaskparser import parser

pagination_args = {
    'offset': wfields.Integer(missing=0),
    'limit': wfields.Integer(missing=20)
}

def paginated(basequery, schema_type, offset=None, limit=None):
    if offset is None or limit is None:
        args = parser.parse(pagination_args, request)
        if offset is None:
            offset = args['offset']

        if limit is None:
            limit = args['limit']

    data = {
        'offset': offset,
        'limit': limit,
        'items': basequery.limit(limit).offset(offset),
        'totalItems': basequery.count()
    }

    class _Pagination(Schema):
        offset = fields.Integer()
        limit = fields.Integer()
        totalItems = fields.Integer()
        items = fields.Nested(schema_type, many=True)

    return _Pagination().dump(data)

__all__ = ['paginated']


from webargs.flaskparser import FlaskParser
from webargs.core import ValidationError
from flask import jsonify


class BetterFlaskParser(FlaskParser):
    def handle_error(self, error):
        """
        Don't raise a HTTPException via abort. Instead we will throw the 
        ValidationError and handle it with our flask Exception handler.

        This allows a common code path for both Flask-Restful AND standard 
        Flask Views.
        """
        raise error

parser = BetterFlaskParser()
use_args = parser.use_args
use_kwargs = parser.use_kwargs

def handle_validation_error(exc):
    """
    If using the above parser, if an exception occurs, it will throw the 
    original ValidationError. This circumvents the capture inside Flask-Restful
    (if it is being used at all). 

    Instead of capturing all 422 HTTPExceptions, you register this error 
    handler with ValidationError:

    ```
    app.errorhandler(ValidationError)(handle_validation_error)
    ```

    This function will produce a jsonified response with the field errors from
    the ValidationError.

    Warning: This handler is incompatible with the standard FlaskParser, since
    it throws HTTPExceptions (via abort). Registering this handler with 
    errorhandler(422) will not work.

    """

    assert(type(exc) == ValidationError)
    return jsonify(exc.messages), exc.status_code

__all__ = ['BetterFlaskParser', 'parser', 'use_args', 'use_kwargs',
           'handle_validation_error']




from setuptools import setup
import os

PACKAGE_PATH = './twopi_flask_utils'

packages = []

for item in os.listdir(PACKAGE_PATH):
    path = os.path.join(PACKAGE_PATH, item)
    if not item.startswith('_') and os.path.isdir(path):
        packages.append(('twopi_flask_utils.' + item, path))

extras = {
    'ldap': [],
    'restful': ['flask-restful'],
    'celery': ['celery'],
    'sentry': ['raven[flask]'],
    'pagination': ['webargs', 'marshmallow'],
    'webargs': ['webargs']
}


tests_require = []
for x in extras.values():
    tests_require.extend(x)


def get_version():
    version_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'VERSION')
    v = open(version_path).read()
    if type(v) == str:
        return v.strip()
    return v.decode('UTF-8').strip()


readme_path = os.path.join(os.path.dirname(
  os.path.abspath(__file__)),
  'README.rst',
)
long_description = open(readme_path).read()

try:
    version = get_version()
except Exception as e:
    version = '0.0.0-dev'

setup(
    name='twopi-flask-utils',
    version=version,
    package_dir=dict(packages),
    packages=list(map(lambda x: x[0], packages)),
    extras_require=extras,
    tests_require=tests_require,
    test_suite='tests',
    url='https://github.com/TwoPiCode/twopi-flask-utils',
    author='Nick Whyte',
    author_email='nick@twopicode.com',
    description=('A set of utilities to make working with flask web '
                 'applications easier.'),
    long_description=long_description
)
